@@ -368,7 +368,7 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
         }
 
         Ok(EntryPointBinding {
-            arg_name: self.namer.call_unique(struct_name.to_lowercase().as_str()),
+            arg_name: self.namer.call(struct_name.to_lowercase().as_str()),
             ty_name: struct_name,
             members,
         })
@@ -391,14 +391,10 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
             match module.types[arg.ty].inner {
                 TypeInner::Struct { ref members, .. } => {
                     for member in members.iter() {
-                        let member_name = if let Some(ref name) = member.name {
-                            name
-                        } else {
-                            "member"
-                        };
+                        let name = self.namer.call_or(&member.name, "member");
                         let index = fake_members.len() as u32;
                         fake_members.push(EpStructMember {
-                            name: self.namer.call(member_name),
+                            name,
                             ty: member.ty,
                             binding: member.binding.clone(),
                             index,
@@ -406,11 +402,7 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
                     }
                 }
                 _ => {
-                    let member_name = if let Some(ref name) = arg.name {
-                        self.namer.call_unique(name)
-                    } else {
-                        self.namer.call("member")
-                    };
+                    let member_name = self.namer.call_or(&arg.name, "member");
                     let index = fake_members.len() as u32;
                     fake_members.push(EpStructMember {
                         name: member_name,
@@ -448,11 +440,7 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
         };
 
         for member in members.iter() {
-            let member_name = if let Some(ref name) = member.name {
-                self.namer.call_unique(name)
-            } else {
-                self.namer.call("member")
-            };
+            let member_name = self.namer.call_or(&member.name, "member");
             let index = fake_members.len() as u32;
             fake_members.push(EpStructMember {
                 name: member_name,
@@ -1057,15 +1045,22 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
         match *stmt {
             Statement::Emit(ref range) => {
                 for handle in range.clone() {
-                    let expr_name = if let Some(name) = func_ctx.named_expressions.get(&handle) {
+                    let info = &func_ctx.info[handle];
+                    let ptr_class = info.ty.inner_with(&module.types).pointer_class();
+                    let expr_name = if ptr_class.is_some() {
+                        // HLSL can't save a pointer-valued expression in a variable,
+                        // but we shouldn't ever need to: they should never be named expressions,
+                        // and none of the expression types flagged by bake_ref_count can be pointer-valued.
+                        None
+                    } else if let Some(name) = func_ctx.named_expressions.get(&handle) {
                         // Front end provides names for all variables at the start of writing.
                         // But we write them to step by step. We need to recache them
                         // Otherwise, we could accidentally write variable name instead of full expression.
                         // Also, we use sanitized names! It defense backend from generating variable with name from reserved keywords.
-                        Some(self.namer.call_unique(name))
+                        Some(self.namer.call(name))
                     } else {
                         let min_ref_count = func_ctx.expressions[handle].bake_ref_count();
-                        if min_ref_count <= func_ctx.info[handle].ref_count {
+                        if min_ref_count <= info.ref_count {
                             Some(format!("_expr{}", handle.index()))
                         } else {
                             None
@@ -1152,7 +1147,7 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
                     };
                     let final_name = match ep_output {
                         Some(ep_output) => {
-                            let final_name = self.namer.call_unique(&variable_name);
+                            let final_name = self.namer.call(&variable_name);
                             write!(
                                 self.out,
                                 "{}const {} {} = {{ ",
@@ -1343,10 +1338,18 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
                     res_name, var_name, fun_str
                 )?;
                 self.write_storage_address(module, &chain, func_ctx)?;
-                if let crate::AtomicFunction::Exchange { compare: Some(_) } = *fun {
-                    return Err(Error::Unimplemented("atomic CompareExchange".to_string()));
-                }
                 write!(self.out, ", ")?;
+                // handle the special cases
+                match *fun {
+                    crate::AtomicFunction::Subtract => {
+                        // we just wrote `InterlockedAdd`, so negate the argument
+                        write!(self.out, "-")?;
+                    }
+                    crate::AtomicFunction::Exchange { compare: Some(_) } => {
+                        return Err(Error::Unimplemented("atomic CompareExchange".to_string()));
+                    }
+                    _ => {}
+                }
                 self.write_expr(module, value, func_ctx)?;
                 writeln!(self.out, ", {});", res_name)?;
                 self.temp_access_chain = chain;
@@ -1362,13 +1365,24 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
                 write!(self.out, "switch(")?;
                 self.write_expr(module, selector, func_ctx)?;
                 writeln!(self.out, ") {{")?;
+                let type_postfix = match *func_ctx.info[selector].ty.inner_with(&module.types) {
+                    crate::TypeInner::Scalar {
+                        kind: crate::ScalarKind::Uint,
+                        ..
+                    } => "u",
+                    _ => "",
+                };
 
                 // Write all cases
                 let indent_level_1 = level.next();
                 let indent_level_2 = indent_level_1.next();
 
                 for case in cases {
-                    writeln!(self.out, "{}case {}: {{", indent_level_1, case.value)?;
+                    writeln!(
+                        self.out,
+                        "{}case {}{}: {{",
+                        indent_level_1, case.value, type_postfix
+                    )?;
 
                     if case.fall_through {
                         // Generate each fallthrough case statement in a new block. This is done to
@@ -1772,20 +1786,15 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
                 write!(self.out, "{}", op_str)?;
                 self.write_expr(module, expr, func_ctx)?;
             }
-            Expression::As { expr, kind, .. } => {
+            Expression::As {
+                expr,
+                kind,
+                convert,
+            } => {
                 let inner = func_ctx.info[expr].ty.inner_with(&module.types);
-                match *inner {
-                    TypeInner::Vector { size, width, .. } => {
-                        write!(
-                            self.out,
-                            "{}{}",
-                            kind.to_hlsl_str(width)?,
-                            back::vector_size_str(size),
-                        )?;
-                    }
-                    TypeInner::Scalar { width, .. } => {
-                        write!(self.out, "{}", kind.to_hlsl_str(width)?)?
-                    }
+                let (size_str, src_width) = match *inner {
+                    TypeInner::Vector { size, width, .. } => (back::vector_size_str(size), width),
+                    TypeInner::Scalar { width, .. } => ("", width),
                     _ => {
                         return Err(Error::Unimplemented(format!(
                             "write_expr expression::as {:?}",
@@ -1793,7 +1802,8 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
                         )));
                     }
                 };
-                write!(self.out, "(")?;
+                let kind_str = kind.to_hlsl_str(convert.unwrap_or(src_width))?;
+                write!(self.out, "{}{}(", kind_str, size_str,)?;
                 self.write_expr(module, expr, func_ctx)?;
                 write!(self.out, ")")?;
             }
@@ -1802,79 +1812,116 @@ impl<'a, W: fmt::Write> super::Writer<'a, W> {
                 arg,
                 arg1,
                 arg2,
+                arg3,
             } => {
                 use crate::MathFunction as Mf;
 
-                let fun_name = match fun {
+                enum Function {
+                    Asincosh { is_sin: bool },
+                    Atanh,
+                    Regular(&'static str),
+                }
+
+                let fun = match fun {
                     // comparison
-                    Mf::Abs => "abs",
-                    Mf::Min => "min",
-                    Mf::Max => "max",
-                    Mf::Clamp => "clamp",
+                    Mf::Abs => Function::Regular("abs"),
+                    Mf::Min => Function::Regular("min"),
+                    Mf::Max => Function::Regular("max"),
+                    Mf::Clamp => Function::Regular("clamp"),
                     // trigonometry
-                    Mf::Cos => "cos",
-                    Mf::Cosh => "cosh",
-                    Mf::Sin => "sin",
-                    Mf::Sinh => "sinh",
-                    Mf::Tan => "tan",
-                    Mf::Tanh => "tanh",
-                    Mf::Acos => "acos",
-                    Mf::Asin => "asin",
-                    Mf::Atan => "atan",
-                    Mf::Atan2 => "atan2",
+                    Mf::Cos => Function::Regular("cos"),
+                    Mf::Cosh => Function::Regular("cosh"),
+                    Mf::Sin => Function::Regular("sin"),
+                    Mf::Sinh => Function::Regular("sinh"),
+                    Mf::Tan => Function::Regular("tan"),
+                    Mf::Tanh => Function::Regular("tanh"),
+                    Mf::Acos => Function::Regular("acos"),
+                    Mf::Asin => Function::Regular("asin"),
+                    Mf::Atan => Function::Regular("atan"),
+                    Mf::Atan2 => Function::Regular("atan2"),
+                    Mf::Asinh => Function::Asincosh { is_sin: true },
+                    Mf::Acosh => Function::Asincosh { is_sin: false },
+                    Mf::Atanh => Function::Atanh,
                     // decomposition
-                    Mf::Ceil => "ceil",
-                    Mf::Floor => "floor",
-                    Mf::Round => "round",
-                    Mf::Fract => "frac",
-                    Mf::Trunc => "trunc",
-                    Mf::Modf => "modf",
-                    Mf::Frexp => "frexp",
-                    Mf::Ldexp => "ldexp",
+                    Mf::Ceil => Function::Regular("ceil"),
+                    Mf::Floor => Function::Regular("floor"),
+                    Mf::Round => Function::Regular("round"),
+                    Mf::Fract => Function::Regular("frac"),
+                    Mf::Trunc => Function::Regular("trunc"),
+                    Mf::Modf => Function::Regular("modf"),
+                    Mf::Frexp => Function::Regular("frexp"),
+                    Mf::Ldexp => Function::Regular("ldexp"),
                     // exponent
-                    Mf::Exp => "exp",
-                    Mf::Exp2 => "exp2",
-                    Mf::Log => "log",
-                    Mf::Log2 => "log2",
-                    Mf::Pow => "pow",
+                    Mf::Exp => Function::Regular("exp"),
+                    Mf::Exp2 => Function::Regular("exp2"),
+                    Mf::Log => Function::Regular("log"),
+                    Mf::Log2 => Function::Regular("log2"),
+                    Mf::Pow => Function::Regular("pow"),
                     // geometry
-                    Mf::Dot => "dot",
+                    Mf::Dot => Function::Regular("dot"),
                     //Mf::Outer => ,
-                    Mf::Cross => "cross",
-                    Mf::Distance => "distance",
-                    Mf::Length => "length",
-                    Mf::Normalize => "normalize",
-                    Mf::FaceForward => "faceforward",
-                    Mf::Reflect => "reflect",
-                    Mf::Refract => "refract",
+                    Mf::Cross => Function::Regular("cross"),
+                    Mf::Distance => Function::Regular("distance"),
+                    Mf::Length => Function::Regular("length"),
+                    Mf::Normalize => Function::Regular("normalize"),
+                    Mf::FaceForward => Function::Regular("faceforward"),
+                    Mf::Reflect => Function::Regular("reflect"),
+                    Mf::Refract => Function::Regular("refract"),
                     // computational
-                    Mf::Sign => "sign",
-                    Mf::Fma => "fma",
-                    Mf::Mix => "lerp",
-                    Mf::Step => "step",
-                    Mf::SmoothStep => "smoothstep",
-                    Mf::Sqrt => "sqrt",
-                    Mf::InverseSqrt => "rsqrt",
+                    Mf::Sign => Function::Regular("sign"),
+                    Mf::Fma => Function::Regular("fma"),
+                    Mf::Mix => Function::Regular("lerp"),
+                    Mf::Step => Function::Regular("step"),
+                    Mf::SmoothStep => Function::Regular("smoothstep"),
+                    Mf::Sqrt => Function::Regular("sqrt"),
+                    Mf::InverseSqrt => Function::Regular("rsqrt"),
                     //Mf::Inverse =>,
-                    Mf::Transpose => "transpose",
-                    Mf::Determinant => "determinant",
+                    Mf::Transpose => Function::Regular("transpose"),
+                    Mf::Determinant => Function::Regular("determinant"),
                     // bits
-                    Mf::CountOneBits => "countbits",
-                    Mf::ReverseBits => "reversebits",
+                    Mf::CountOneBits => Function::Regular("countbits"),
+                    Mf::ReverseBits => Function::Regular("reversebits"),
                     _ => return Err(Error::Unimplemented(format!("write_expr_math {:?}", fun))),
                 };
 
-                write!(self.out, "{}(", fun_name)?;
-                self.write_expr(module, arg, func_ctx)?;
-                if let Some(arg) = arg1 {
-                    write!(self.out, ", ")?;
-                    self.write_expr(module, arg, func_ctx)?;
+                match fun {
+                    Function::Asincosh { is_sin } => {
+                        write!(self.out, "log(")?;
+                        self.write_expr(module, arg, func_ctx)?;
+                        write!(self.out, " + sqrt(")?;
+                        self.write_expr(module, arg, func_ctx)?;
+                        write!(self.out, " * ")?;
+                        self.write_expr(module, arg, func_ctx)?;
+                        match is_sin {
+                            true => write!(self.out, " + 1.0))")?,
+                            false => write!(self.out, " - 1.0))")?,
+                        }
+                    }
+                    Function::Atanh => {
+                        write!(self.out, "0.5 * log((1.0 + ")?;
+                        self.write_expr(module, arg, func_ctx)?;
+                        write!(self.out, ") / (1.0 - ")?;
+                        self.write_expr(module, arg, func_ctx)?;
+                        write!(self.out, "))")?;
+                    }
+                    Function::Regular(fun_name) => {
+                        write!(self.out, "{}(", fun_name)?;
+                        self.write_expr(module, arg, func_ctx)?;
+                        if let Some(arg) = arg1 {
+                            write!(self.out, ", ")?;
+                            self.write_expr(module, arg, func_ctx)?;
+                        }
+                        if let Some(arg) = arg2 {
+                            write!(self.out, ", ")?;
+                            self.write_expr(module, arg, func_ctx)?;
+                        }
+                        if let Some(arg) = arg3 {
+                            write!(self.out, ", ")?;
+                            self.write_expr(module, arg, func_ctx)?;
+                        }
+                        write!(self.out, ")")?
+                    }
                 }
-                if let Some(arg) = arg2 {
-                    write!(self.out, ", ")?;
-                    self.write_expr(module, arg, func_ctx)?;
-                }
-                write!(self.out, ")")?
             }
             Expression::Swizzle {
                 size,
