@@ -5,6 +5,8 @@ use thiserror::Error;
 
 pub use writer::Writer;
 
+const BAKE_PREFIX: &str = "e";
+
 #[derive(Error, Debug)]
 pub enum Error {
     #[error(transparent)]
@@ -31,6 +33,7 @@ impl crate::AtomicFunction {
     fn to_wgsl(self) -> &'static str {
         match self {
             Self::Add => "Add",
+            Self::Subtract => "Sub",
             Self::And => "And",
             Self::InclusiveOr => "Or",
             Self::ExclusiveOr => "Xor",
