@@ -1,15 +1,16 @@
 use super::{
     ast::*,
     builtins::{inject_builtin, inject_double_builtin, sampled_to_depth},
-    context::{Context, StmtContext},
+    context::{Context, ExprPos, StmtContext},
     error::{Error, ErrorKind},
     types::scalar_components,
-    Parser, Result, SourceMetadata,
+    Parser, Result,
 };
 use crate::{
     front::glsl::types::type_power, proc::ensure_block_returns, Arena, Block, Constant,
     ConstantInner, EntryPoint, Expression, FastHashMap, Function, FunctionArgument, FunctionResult,
-    Handle, LocalVariable, ScalarKind, ScalarValue, Span, Statement, StructMember, Type, TypeInner,
+    Handle, LocalVariable, ScalarKind, ScalarValue, Span, Statement, StorageClass, StructMember,
+    Type, TypeInner,
 };
 use std::iter;
 
@@ -18,7 +19,7 @@ impl Parser {
         &mut self,
         scalar_kind: ScalarKind,
         value: u64,
-        meta: SourceMetadata,
+        meta: Span,
     ) -> Handle<Constant> {
         let value = match scalar_kind {
             ScalarKind::Uint => ScalarValue::Uint(value),
@@ -33,7 +34,7 @@ impl Parser {
                 specialization: None,
                 inner: ConstantInner::Scalar { width: 4, value },
             },
-            meta.as_span(),
+            meta,
         )
     }
 
@@ -44,11 +45,11 @@ impl Parser {
         body: &mut Block,
         fc: FunctionCallKind,
         raw_args: &[Handle<HirExpr>],
-        meta: SourceMetadata,
+        meta: Span,
     ) -> Result<Option<Handle<Expression>>> {
         let args: Vec<_> = raw_args
             .iter()
-            .map(|e| ctx.lower_expect_inner(stmt, self, *e, false, body))
+            .map(|e| ctx.lower_expect_inner(stmt, self, *e, ExprPos::Rhs, body))
             .collect::<Result<_>>()?;
 
         match fc {
@@ -149,7 +150,7 @@ impl Parser {
                                     // value is used to initialize all the values along the diagonal of
                                     // the matrix; the rest are given zeros.
                                     let mut components = Vec::with_capacity(columns as usize);
-                                    let vector_ty = self.module.types.fetch_or_append(
+                                    let vector_ty = self.module.types.insert(
                                         Type {
                                             name: None,
                                             inner: TypeInner::Vector {
@@ -158,7 +159,7 @@ impl Parser {
                                                 width,
                                             },
                                         },
-                                        meta.as_span(),
+                                        meta,
                                     );
                                     let zero_constant = self.module.constants.fetch_or_append(
                                         Constant {
@@ -169,7 +170,7 @@ impl Parser {
                                                 value: ScalarValue::Float(0.0),
                                             },
                                         },
-                                        meta.as_span(),
+                                        meta,
                                     );
                                     let zero = ctx.add_expression(
                                         Expression::Constant(zero_constant),
@@ -243,7 +244,7 @@ impl Parser {
                                 }
                             }
                         }
-                        TypeInner::Struct { .. } => ctx.add_expression(
+                        TypeInner::Struct { .. } | TypeInner::Array { .. } => ctx.add_expression(
                             Expression::Compose {
                                 ty,
                                 components: args.into_iter().map(|arg| arg.0).collect(),
@@ -253,7 +254,7 @@ impl Parser {
                         ),
                         _ => {
                             self.errors.push(Error {
-                                kind: ErrorKind::SemanticError("Bad cast".into()),
+                                kind: ErrorKind::SemanticError("Bad type constructor".into()),
                                 meta,
                             });
 
@@ -296,7 +297,7 @@ impl Parser {
                                 }
                             }
 
-                            let ty = self.module.types.fetch_or_append(
+                            let ty = self.module.types.insert(
                                 Type {
                                     name: None,
                                     inner: TypeInner::Vector {
@@ -305,7 +306,7 @@ impl Parser {
                                         width,
                                     },
                                 },
-                                meta.as_span(),
+                                meta,
                             );
 
                             for chunk in flattened.chunks(rows as usize) {
@@ -350,9 +351,9 @@ impl Parser {
         stmt: &StmtContext,
         body: &mut Block,
         name: String,
-        args: Vec<(Handle<Expression>, SourceMetadata)>,
+        args: Vec<(Handle<Expression>, Span)>,
         raw_args: &[Handle<HirExpr>],
-        meta: SourceMetadata,
+        meta: Span,
     ) -> Result<Option<Handle<Expression>>> {
         // If the name for the function hasn't yet been initialized check if any
         // builtin can be injected.
@@ -376,128 +377,103 @@ impl Parser {
         // Borrow again but without mutability
         let declaration = self.lookup_function.get(&name).unwrap();
 
-        // Helper enum containing the type of conversion need for a call
-        #[derive(PartialEq, Eq, Clone, Copy, Debug)]
-        enum Conversion {
-            // No conversion needed
-            Exact,
-            // Float to double conversion needed
-            FloatToDouble,
-            // Int or uint to float conversion needed
-            IntToFloat,
-            // Int or uint to double conversion needed
-            IntToDouble,
-            // Other type of conversion needed
-            Other,
-            // No conversion was yet registered
-            None,
-        }
-
-        let mut maybe_decl = None;
+        // Possibly contains the overload to be used in the call
+        let mut maybe_overload = None;
+        // The conversions needed for the best analyzed overload, this is initialized all to
+        // `NONE` to make sure that conversions always pass the first time without ambiguity
         let mut old_conversions = vec![Conversion::None; args.len()];
+        // Tracks whether the comparison between overloads lead to an ambiguity
         let mut ambiguous = false;
 
-        'outer: for decl in declaration.overloads.iter() {
-            if args.len() != decl.parameters.len() {
+        // Iterate over all the available overloads to select either an exact match or a
+        // overload which has suitable implicit conversions
+        'outer: for overload in declaration.overloads.iter() {
+            // If the overload and the function call don't have the same number of arguments
+            // continue to the next overload
+            if args.len() != overload.parameters.len() {
                 continue;
             }
 
+            // Stores whether the current overload matches exactly the function call
             let mut exact = true;
             // State of the selection
-            // If None we still don't know what is the best declaration
-            // If Some(true) the new declaration is better
-            // If Some(false) the old declaration is better
+            // If None we still don't know what is the best overload
+            // If Some(true) the new overload is better
+            // If Some(false) the old overload is better
             let mut superior = None;
+            // Store the conversions for the current overload so that later they can replace the
+            // conversions used for querying the best overload
             let mut new_conversions = vec![Conversion::None; args.len()];
 
-            for ((i, decl_arg), call_arg) in decl.parameters.iter().enumerate().zip(args.iter()) {
-                use ScalarKind::*;
+            // Loop trough the overload parameters and check if the current overload is better
+            // compared to the previous best overload.
+            for (i, overload_parameter) in overload.parameters.iter().enumerate() {
+                let call_argument = &args[i];
+                let parameter_info = &overload.parameters_info[i];
 
-                if decl.parameters_info[i].depth {
+                // If the image is used in the overload as a depth texture convert it
+                // before comparing, otherwise exact matches wouldn't be reported
+                if parameter_info.depth {
                     sampled_to_depth(
                         &mut self.module,
                         ctx,
-                        call_arg.0,
-                        call_arg.1,
+                        call_argument.0,
+                        call_argument.1,
                         &mut self.errors,
-                    )?;
-                    self.invalidate_expression(ctx, call_arg.0, call_arg.1)?
+                    );
+                    self.invalidate_expression(ctx, call_argument.0, call_argument.1)?
                 }
 
-                let decl_inner = &self.module.types[*decl_arg].inner;
-                let call_inner = self.resolve_type(ctx, call_arg.0, call_arg.1)?;
+                let overload_param_ty = &self.module.types[*overload_parameter].inner;
+                let call_arg_ty = self.resolve_type(ctx, call_argument.0, call_argument.1)?;
 
-                if decl_inner == call_inner {
+                // If the types match there's no need to check for conversions so continue
+                if overload_param_ty == call_arg_ty {
                     new_conversions[i] = Conversion::Exact;
                     continue;
                 }
 
-                exact = false;
-
-                let (decl_kind, decl_width, call_kind, call_width) = match (decl_inner, call_inner)
-                {
-                    (
-                        &TypeInner::Scalar {
-                            kind: decl_kind,
-                            width: decl_width,
-                        },
-                        &TypeInner::Scalar {
-                            kind: call_kind,
-                            width: call_width,
-                        },
-                    ) => (decl_kind, decl_width, call_kind, call_width),
-                    (
-                        &TypeInner::Vector {
-                            kind: decl_kind,
-                            size: decl_size,
-                            width: decl_width,
-                        },
-                        &TypeInner::Vector {
-                            kind: call_kind,
-                            size: call_size,
-                            width: call_width,
-                        },
-                    ) if decl_size == call_size => (decl_kind, decl_width, call_kind, call_width),
-                    (
-                        &TypeInner::Matrix {
-                            rows: decl_rows,
-                            columns: decl_columns,
-                            width: decl_width,
-                        },
-                        &TypeInner::Matrix {
-                            rows: call_rows,
-                            columns: call_columns,
-                            width: call_width,
-                        },
-                    ) if decl_columns == call_columns && decl_rows == call_rows => {
-                        (Float, decl_width, Float, call_width)
-                    }
-                    _ => continue 'outer,
-                };
-
-                if type_power(decl_kind, decl_width) < type_power(call_kind, call_width) {
+                // If the argument is to be passed as a pointer (i.e. either `out` or
+                // `inout` where used as qualifiers) no conversion shall be performed
+                if parameter_info.qualifier.is_lhs() {
                     continue 'outer;
                 }
 
-                let conversion = match ((decl_kind, decl_width), (call_kind, call_width)) {
-                    ((Float, 8), (Float, 4)) => Conversion::FloatToDouble,
-                    ((Float, 4), (Sint, _)) | ((Float, 4), (Uint, _)) => Conversion::IntToFloat,
-                    ((Float, 8), (Sint, _)) | ((Float, 8), (Uint, _)) => Conversion::IntToDouble,
-                    _ => Conversion::Other,
+                // Try to get the type of conversion needed otherwise this overload can't be used
+                // since no conversion makes it possible so skip it
+                let conversion = match conversion(overload_param_ty, call_arg_ty) {
+                    Some(info) => info,
+                    None => continue 'outer,
                 };
 
-                // true - New declaration argument has a better conversion
-                // false - Old declaration argument has a better conversion
+                // At this point a conversion will be needed so the overload no longer
+                // exactly matches the call arguments
+                exact = false;
+
+                // Compare the conversions needed for this overload parameter to that of the
+                // last overload analyzed respective parameter, the value is:
+                // - `true` when the new overload argument has a better conversion
+                // - `false` when the old overload argument has a better conversion
                 let best_arg = match (conversion, old_conversions[i]) {
+                    // An exact match is always better, we don't need to check this for the
+                    // current overload since it was checked earlier
                     (_, Conversion::Exact) => false,
-                    (Conversion::FloatToDouble, _)
-                    | (_, Conversion::None)
-                    | (Conversion::IntToFloat, Conversion::IntToDouble) => true,
-                    (_, Conversion::FloatToDouble)
-                    | (Conversion::IntToDouble, Conversion::IntToFloat) => false,
+                    // No overload was yet analyzed so this one is the best yet
+                    (_, Conversion::None) => true,
+                    // A conversion from a float to a double is the best possible conversion
+                    (Conversion::FloatToDouble, _) => true,
+                    (_, Conversion::FloatToDouble) => false,
+                    // A conversion from a float to an integer is preferred than one
+                    // from double to an integer
+                    (Conversion::IntToFloat, Conversion::IntToDouble) => true,
+                    (Conversion::IntToDouble, Conversion::IntToFloat) => false,
+                    // This case handles things like no conversion and exact which were already
+                    // treated and other cases which no conversion is better than the other
                     _ => continue,
                 };
 
+                // Check if the best parameter corresponds to the current selected overload
+                // to pass to the next comparation, if this isn't true mark it as ambiguous
                 match best_arg {
                     true => match superior {
                         Some(false) => ambiguous = true,
@@ -513,28 +489,33 @@ impl Parser {
                 }
             }
 
+            // The overload matches exactly the function call so there's no ambiguity (since
+            // repeated overload aren't allowed) and the current overload is selected, no
+            // further querying is needed.
             if exact {
-                maybe_decl = Some(decl);
+                maybe_overload = Some(overload);
                 ambiguous = false;
                 break;
             }
 
             match superior {
-                // New declaration is better keep it
+                // New overload is better keep it
                 Some(true) => {
-                    maybe_decl = Some(decl);
+                    maybe_overload = Some(overload);
                     // Replace the conversions
                     old_conversions = new_conversions;
                 }
-                // Old declaration is better do nothing
+                // Old overload is better do nothing
                 Some(false) => {}
-                // No declaration was better than the other this can be caused
-                // when
+                // No overload was better than the other this can be caused
+                // when all conversions are ambiguous in which the overloads themselves are
+                // ambiguous.
                 None => {
                     ambiguous = true;
-                    // Assign the new declaration to make sure we always have
-                    // one to make error reporting happy
-                    maybe_decl = Some(decl);
+                    // Assign the new overload, this helps ensures that in this case of
+                    // amiguity the parsing won't end immediately and allow for further
+                    // collection of errors.
+                    maybe_overload = Some(overload);
                 }
             }
         }
@@ -548,66 +529,109 @@ impl Parser {
             })
         }
 
-        let decl = maybe_decl.ok_or_else(|| Error {
+        let overload = maybe_overload.ok_or_else(|| Error {
             kind: ErrorKind::SemanticError(format!("Unknown function '{}'", name).into()),
             meta,
         })?;
 
-        let parameters_info = decl.parameters_info.clone();
-        let parameters = decl.parameters.clone();
-        let is_void = decl.void;
-        let kind = decl.kind;
+        let parameters_info = overload.parameters_info.clone();
+        let parameters = overload.parameters.clone();
+        let is_void = overload.void;
+        let kind = overload.kind;
 
         let mut arguments = Vec::with_capacity(args.len());
         let mut proxy_writes = Vec::new();
+        // Iterate trough the function call arguments applying transformations as needed
         for (parameter_info, (expr, parameter)) in parameters_info
             .iter()
             .zip(raw_args.iter().zip(parameters.iter()))
         {
             let (mut handle, meta) =
-                ctx.lower_expect_inner(stmt, self, *expr, parameter_info.qualifier.is_lhs(), body)?;
+                ctx.lower_expect_inner(stmt, self, *expr, parameter_info.qualifier.as_pos(), body)?;
 
-            if let TypeInner::Vector { size, kind, width } =
-                *self.resolve_type(ctx, handle, meta)?
-            {
-                if parameter_info.qualifier.is_lhs()
-                    && matches!(ctx[handle], Expression::Swizzle { .. })
-                {
-                    let ty = self.module.types.fetch_or_append(
-                        Type {
-                            name: None,
-                            inner: TypeInner::Vector { size, kind, width },
-                        },
-                        Span::Unknown,
-                    );
-                    let temp_var = ctx.locals.append(
-                        LocalVariable {
-                            name: None,
-                            ty,
-                            init: None,
-                        },
-                        Span::Unknown,
-                    );
-                    let temp_expr = ctx.add_expression(
-                        Expression::LocalVariable(temp_var),
-                        SourceMetadata::none(),
-                        body,
-                    );
+            if parameter_info.qualifier.is_lhs() {
+                let (ty, value) = match *self.resolve_type(ctx, handle, meta)? {
+                    // If the argument is to be passed as a pointer but the type of the
+                    // expression returns a vector it must mean that it was for example
+                    // swizzled and it must be spilled into a local before calling
+                    // TODO: this part doesn't work because of #1385 once that's sorted out
+                    // revisit this part.
+                    TypeInner::Vector { size, kind, width } => (
+                        self.module.types.insert(
+                            Type {
+                                name: None,
+                                inner: TypeInner::Vector { size, kind, width },
+                            },
+                            Span::default(),
+                        ),
+                        handle,
+                    ),
+                    // If the argument is a pointer whose storage class isn't `Function` an
+                    // indirection trough a local variable is needed to align the storage
+                    // classes of the call argument and the overload parameter
+                    TypeInner::Pointer { base, class } if class != StorageClass::Function => (
+                        base,
+                        ctx.add_expression(
+                            Expression::Load { pointer: handle },
+                            Span::default(),
+                            body,
+                        ),
+                    ),
+                    TypeInner::ValuePointer {
+                        size,
+                        kind,
+                        width,
+                        class,
+                    } if class != StorageClass::Function => {
+                        let inner = match size {
+                            Some(size) => TypeInner::Vector { size, kind, width },
+                            None => TypeInner::Scalar { kind, width },
+                        };
 
-                    body.push(
-                        Statement::Store {
-                            pointer: temp_expr,
-                            value: handle,
-                        },
-                        Span::Unknown,
-                    );
+                        (
+                            self.module
+                                .types
+                                .insert(Type { name: None, inner }, Span::default()),
+                            ctx.add_expression(
+                                Expression::Load { pointer: handle },
+                                Span::default(),
+                                body,
+                            ),
+                        )
+                    }
+                    _ => {
+                        arguments.push(handle);
+                        continue;
+                    }
+                };
 
-                    arguments.push(temp_expr);
-                    proxy_writes.push((*expr, temp_expr));
-                    continue;
-                }
+                let temp_var = ctx.locals.append(
+                    LocalVariable {
+                        name: None,
+                        ty,
+                        init: None,
+                    },
+                    Span::default(),
+                );
+                let temp_expr =
+                    ctx.add_expression(Expression::LocalVariable(temp_var), Span::default(), body);
+
+                body.push(
+                    Statement::Store {
+                        pointer: temp_expr,
+                        value,
+                    },
+                    Span::default(),
+                );
+
+                arguments.push(temp_expr);
+                // Register the temporary local to be written back to it's original
+                // place after the function call
+                proxy_writes.push((handle, temp_expr));
+                continue;
             }
 
+            // Apply implicit conversions as needed
             let scalar_components = scalar_components(&self.module.types[*parameter].inner);
             if let Some((kind, width)) = scalar_components {
                 ctx.implicit_conversion(self, &mut handle, meta, kind, width)?;
@@ -632,23 +656,25 @@ impl Parser {
                         arguments,
                         result,
                     },
-                    meta.as_span(),
+                    meta,
                 );
 
                 ctx.emit_start();
-                for (tgt, pointer) in proxy_writes {
+
+                // Write back all the variables that were scheduled to their original place
+                for (original, pointer) in proxy_writes {
                     let value = ctx.add_expression(Expression::Load { pointer }, meta, body);
-                    let target = ctx.lower_expect_inner(stmt, self, tgt, true, body)?.0;
 
                     ctx.emit_flush(body);
+                    ctx.emit_start();
+
                     body.push(
                         Statement::Store {
-                            pointer: target,
+                            pointer: original,
                             value,
                         },
-                        meta.as_span(),
+                        meta,
                     );
-                    ctx.emit_start();
                 }
 
                 Ok(result)
@@ -665,7 +691,7 @@ impl Parser {
         name: String,
         result: Option<FunctionResult>,
         mut body: Block,
-        meta: SourceMetadata,
+        meta: Span,
     ) {
         if self.lookup_function.get(&name).is_none() {
             let declaration = self.lookup_function.entry(name.clone()).or_default();
@@ -738,14 +764,14 @@ impl Parser {
             match decl.kind {
                 FunctionKind::Call(handle) => *self.module.functions.get_mut(handle) = function,
                 FunctionKind::Macro(_) => {
-                    let handle = module.functions.append(function, meta.as_span());
+                    let handle = module.functions.append(function, meta);
                     decl.kind = FunctionKind::Call(handle)
                 }
             }
             return;
         }
 
-        let handle = module.functions.append(function, meta.as_span());
+        let handle = module.functions.append(function, meta);
         declaration.overloads.push(Overload {
             parameters,
             parameters_info,
@@ -760,7 +786,7 @@ impl Parser {
         ctx: Context,
         name: String,
         result: Option<FunctionResult>,
-        meta: SourceMetadata,
+        meta: Span,
     ) {
         if self.lookup_function.get(&name).is_none() {
             let declaration = self.lookup_function.entry(name.clone()).or_default();
@@ -820,7 +846,7 @@ impl Parser {
             });
         }
 
-        let handle = module.functions.append(function, meta.as_span());
+        let handle = module.functions.append(function, meta);
         declaration.overloads.push(Overload {
             parameters,
             parameters_info,
@@ -910,7 +936,7 @@ impl Parser {
         }
 
         let (ty, value) = if !components.is_empty() {
-            let ty = self.module.types.append(
+            let ty = self.module.types.insert(
                 Type {
                     name: None,
                     inner: TypeInner::Struct {
@@ -963,3 +989,88 @@ fn is_double(ty: &TypeInner) -> bool {
         _ => false,
     }
 }
+
+/// Helper enum containing the type of conversion need for a call
+#[derive(PartialEq, Eq, Clone, Copy, Debug)]
+enum Conversion {
+    /// No conversion needed
+    Exact,
+    /// Float to double conversion needed
+    FloatToDouble,
+    /// Int or uint to float conversion needed
+    IntToFloat,
+    /// Int or uint to double conversion needed
+    IntToDouble,
+    /// Other type of conversion needed
+    Other,
+    /// No conversion was yet registered
+    None,
+}
+
+/// Helper function, returns the type of conversion from `source` to `target`, if a
+/// conversion is not possible returns None.
+fn conversion(target: &TypeInner, source: &TypeInner) -> Option<Conversion> {
+    use ScalarKind::*;
+
+    // Gather the `ScalarKind` and scalar width from both the target and the source
+    let (target_kind, target_width, source_kind, source_width) = match (target, source) {
+        // Conversions between scalars are allowed
+        (
+            &TypeInner::Scalar {
+                kind: tgt_kind,
+                width: tgt_width,
+            },
+            &TypeInner::Scalar {
+                kind: src_kind,
+                width: src_width,
+            },
+        ) => (tgt_kind, tgt_width, src_kind, src_width),
+        // Conversions between vectors of the same size are allowed
+        (
+            &TypeInner::Vector {
+                kind: tgt_kind,
+                size: tgt_size,
+                width: tgt_width,
+            },
+            &TypeInner::Vector {
+                kind: src_kind,
+                size: src_size,
+                width: src_width,
+            },
+        ) if tgt_size == src_size => (tgt_kind, tgt_width, src_kind, src_width),
+        // Conversions between matrices of the same size are allowed
+        (
+            &TypeInner::Matrix {
+                rows: tgt_rows,
+                columns: tgt_cols,
+                width: tgt_width,
+            },
+            &TypeInner::Matrix {
+                rows: src_rows,
+                columns: src_cols,
+                width: src_width,
+            },
+        ) if tgt_cols == src_cols && tgt_rows == src_rows => (Float, tgt_width, Float, src_width),
+        _ => return None,
+    };
+
+    // Check if source can be converted into target, if this is the case then the type
+    // power of target must be higher than that of source
+    let target_power = type_power(target_kind, target_width);
+    let source_power = type_power(source_kind, source_width);
+    if target_power < source_power {
+        return None;
+    }
+
+    Some(
+        match ((target_kind, target_width), (source_kind, source_width)) {
+            // A conversion from a float to a double is special
+            ((Float, 8), (Float, 4)) => Conversion::FloatToDouble,
+            // A conversion from an integer to a float is special
+            ((Float, 4), (Sint, _)) | ((Float, 4), (Uint, _)) => Conversion::IntToFloat,
+            // A conversion from an integer to a double is special
+            ((Float, 8), (Sint, _)) | ((Float, 8), (Uint, _)) => Conversion::IntToDouble,
+            _ => Conversion::Other,
+        },
+    )
+}
